/*  Convert a custom state machine description into Berkeley's KISS2 format
 *  Copyright (C) 2000  Paul Stoffregen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fsm.h"

struct symbol_struct *sym_list=NULL;
int err_count;

void output_kiss2(void);
void print_all_data(void);
const char *spaces(const char *str, int max);


int main(int argc, char **argv)
{
	if (argc < 2) {
		yyin = stdin;
	} else {
		yyin = fopen(argv[1], "r");
		if (yyin == NULL) {
			fprintf(stderr, "Unable to read %s\n", argv[1]);
			exit(1);
		}
	}
	err_count = 0;
	yyparse();
	if (err_count > 0) {
		fprintf(stderr, "%d problems found in input\n", err_count);
		exit(1);
	}
	/* print_all_data(); */
	output_kiss2();

	return 0;
}


void output_kiss2(void)
{
	struct symbol_struct *p, *state;
	struct output_struct *out;
	struct next_state_struct *ns;
	struct cond_struct *cond;
	char instr[256], outstr[256];
	int n, maxlen=0;

	/* count the number of inputs */
	n = 0;
	for (p = sym_list; p != NULL; p = p->next) {
		if (p->type == TYPE_INPUT) n++;
	}
	printf(".i %d\n", n);

	/* count the number of outputs */
	n = 0;
	for (p = sym_list; p != NULL; p = p->next) {
		if (p->type == TYPE_OUTPUT) n++;
	}
	printf(".o %d\n", n);

	for (p = sym_list; p != NULL; p = p->next) {
		if (p->type == TYPE_STATE && strlen(p->name) > maxlen)
			maxlen = strlen(p->name);
	}

	/* do each state */
	for (state = sym_list; state != NULL; state = state->next) {
		if (state->type != TYPE_STATE) continue;

		/* init each output with its default value */
		for (p = sym_list; p != NULL; p = p->next) {
			if (p->type != TYPE_OUTPUT) continue;
			p->output_value = p->output_default;
		}

		/* now apply the outputs specific for this state */
		for (out = state->outputs; out != NULL; out = out->next) {
			out->sym->output_value = out->value;
		}

		/* create the output string... a moore machine always  */
		/* has the same outputs within a state, regardless of */
		/* the input signals.  mealey machines are not supported */
		n = 0;
		for (p = sym_list; p != NULL; p = p->next) {
			if (p->type != TYPE_OUTPUT) continue;
			if (p->output_value == 0 || p->output_value == 1) {
				outstr[n] = p->output_value + '0';
			} else {
				outstr[n] = '-';
			}
			if (++n >= sizeof(outstr)) break;
		}
		outstr[n] = '\0';

		for (ns = state->next_state; ns != NULL; ns = ns->next) {
			for (p = sym_list; p != NULL; p = p->next) {
				if (p->type != TYPE_INPUT) continue;
				p->input_value = -1;
			}
			for (cond = ns->cond; cond != NULL; cond = cond->next) {
				cond->sym->input_value = cond->value;
			}
			n = 0;
			for (p = sym_list; p != NULL; p = p->next) {
				if (p->type != TYPE_INPUT) continue;
				if (p->input_value == 0 || p->input_value == 1) {
					instr[n] = p->input_value + '0';
				} else {
					instr[n] = '-';
				}
				if (++n >= sizeof(instr)) break;
			}
			instr[n] = '\0';

			printf("%s    %s", instr, state->name);
			printf("%s", spaces(state->name, maxlen + 4));
			printf("%s", ns->next_state->name);
			printf("%s", spaces(ns->next_state->name, maxlen + 4));
			printf("%s\n", outstr);
		}
	}
}


const char *spaces(const char *str, int max)
{
	static char buf[256];
	int i, n;

	n = max - strlen(str);
	if (n < 0) n = 0;
	if (n >= sizeof(buf)) n = sizeof(buf) - 1;
	for (i=0; i<n; i++) buf[i] = ' ';
	buf[i] = '\0';
	return buf;
}




void print_all_data(void)
{
	struct symbol_struct *p;
	struct output_struct *out;
	struct next_state_struct *ns;
	struct cond_struct *cond;
	int first;

	printf("Inputs:\n");
	for (p = sym_list; p != NULL; p = p->next) {
		if (p->type == TYPE_INPUT) {
			printf("  %s\n", p->name);
		}
	}
	printf("Outputs:\n");
	for (p = sym_list; p != NULL; p = p->next) {
		if (p->type == TYPE_OUTPUT) {
			printf("  %s\t\t%d\n", p->name, p->output_default);
		}
	}
	printf("States:\n");
	for (p = sym_list; p != NULL; p = p->next) {
		if (p->type == TYPE_STATE) {
			printf("  %s\n", p->name);
			for (out = p->outputs; out != NULL; out = out->next) {
				printf("    %s = %d\n", out->sym->name, out->value);
			}
			for (ns = p->next_state; ns != NULL; ns = ns->next) {
				printf("    next_state is %s", ns->next_state->name);
				first = 1;
				for (cond = ns->cond; cond != NULL; cond = cond->next) {
					printf("%s", (first) ? "\tIF " : " AND ");
					first = 0;
					printf("%s == %d", cond->sym->name, cond->value);
				}
				printf("\n");
			}
		}
	}
}





struct symbol_struct * add_symbol(char *name, int type, int default_output)
{
	struct symbol_struct *new, *p;

	new = lookup_symbol(name);
	if (new != NULL) {
		fprintf(stderr, "Warning, \"%s\" multiple definitions", name);
		if (new->lineno == lineno) {
			fprintf(stderr, ", line %d\n", lineno);
		} else {
			fprintf(stderr, ", line %d and %d\n", new->lineno, lineno);
		}
		err_count++;
	}
	new = (struct symbol_struct *)malloc(sizeof(struct symbol_struct));
	new->name = name;
	new->type = type;
	new->output_default = default_output;
	new->output_value = -1;
	new->lineno = lineno;
	new->outputs = NULL;
	new->next = NULL;
	if (sym_list == NULL) {
		sym_list = new;
	} else {
		for (p = sym_list; p->next != NULL; p = p->next)  ;
		p->next = new;
	}
	return new;
}

struct symbol_struct * lookup_symbol(const char *name)
{
	struct symbol_struct *p;
	for (p = sym_list; p != NULL; p = p->next) {
		if (strcasecmp(name, p->name) == 0) return p;
	}
	return NULL;
}


void add_output(struct symbol_struct *state, const char *name, int value)
{
	struct output_struct *new;

	/* make sure we got a valid state for this output definition */
	if (state == NULL) {
		fprintf(stderr, "Error, line %d, output \"%s\" must be assigned within a state\n",
			lineno, name);
		err_count++;
		return;
	}
	/* check if this output is already assigned a value */
	for (new = state->outputs; new != NULL; new = new->next) {
		if (strcasecmp(new->sym->name, name) == 0) {
			fprintf(stderr,
				"Warning, line %d, output value \"%s\" redefined in state \"%s\"\n",
				lineno, name, state->name);
			new->value = value;
			return;
		}
	}
	/* add this output assignment to the state */
	new = (struct output_struct *)malloc(sizeof(struct output_struct));
	new->sym = lookup_symbol(name);
	if (new->sym == NULL) {
		fprintf(stderr, "Error, line %d, output \"%s\" undefined\n", lineno, name);
		err_count++;
		free(new);
		return;
	}
	if (new->sym->type != TYPE_OUTPUT) {
		fprintf(stderr, "Error, line %d, assignment to \"%s\" (not an output)\n",
			lineno, name);
		err_count++;
	}
	new->value = value;
	new->next = state->outputs;
	state->outputs = new;
}

void add_next_state(struct symbol_struct *state,
	const char *next_state_name, struct cond_struct *conditions)
{
	struct next_state_struct *new;

	if (state == NULL) {
		fprintf(stderr, "Error, line %d, next state \"%s\" must be assigned within a state\n",
			lineno, next_state_name);
		err_count++;
		return;
	}
	new = (struct next_state_struct *)malloc(sizeof(struct next_state_struct));
	new->sym = state;
	new->next_state = lookup_symbol(next_state_name);
	if (new->next_state == NULL) {
		fprintf(stderr, "Error, line %d, next state \"%s\" undefined\n",
			lineno, next_state_name);
		err_count++;
		free(new);
		return;
	}
	if (new->next_state->type != TYPE_STATE) {
		fprintf(stderr, "Error, line %d, next state \"%s\" isn't a state\n",
			lineno, next_state_name);
		err_count++;
		free(new);
		return;
	}
	new->cond = conditions;
	new->next = state->next_state;
	state->next_state = new;
}





