%{

/*  Parse a custom state machine description language
 *  Copyright (C) 2000  Paul Stoffregen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fsm.h"

int yyerror(char *s);
void undef_input(void);
void undef_output(void);
void undef_state(void);
void err_redefine(void);
void add_symbol_group(int type, int value);
void build_cond_list(const char *name, int value);


struct symbol_struct *current_state=NULL;
struct cond_struct *current_cond_list=NULL;
char *sym_group[256];
int sym_count=0;
int i;

%}

%token STATE NEXT IF NUM SYM DEFAULT
%token INPUT_SIGNALS OUTPUT_SIGNALS STATE_NAMES
%token SYM_STATE SYM_IN SYM_OUT

%%

all: 
	  block
	| block all

block:	  decl_in
	| decl_out
	| decl_state
	| state

decl_in:
	INPUT_SIGNALS ':' sym_group ';' {
		add_symbol_group(TYPE_INPUT, 0);
		/*
		for (i=0; i<sym_count; i++) {
			add_symbol(sym_group[i], TYPE_INPUT, 0);
		}
		sym_count = 0;
		*/
	}

decl_out:
	OUTPUT_SIGNALS ':' sym_group DEFAULT '=' NUM ';' {
		add_symbol_group(TYPE_OUTPUT, $6);
		/*
		for (i=0; i<sym_count; i++) {
			add_symbol(sym_group[i], TYPE_OUTPUT, $6);
		}
		sym_count = 0;
		*/
	}

decl_state:
	STATE_NAMES ':' sym_group ';' {
		add_symbol_group(TYPE_STATE, 0);
		/*
		for (i=0; i<sym_count; i++) {
			add_symbol(sym_group[i], TYPE_STATE, 0);
		}
		sym_count = 0;
		*/
	}

sym_group:
	  symbol
	| symbol ',' sym_group

symbol:
	  SYM {sym_group[sym_count++] = strdup(lex_sym_name);}
	| SYM_IN {err_redefine();}
	| SYM_OUT {err_redefine();}
	| SYM_STATE {err_redefine();}

state:
	  STATE sym_state ':' actions
	| STATE SYM ':' actions {undef_state();}

sym_state:
	SYM_STATE {current_state = lookup_symbol(lex_sym_name);}

actions:
	  action
	| action actions

action:
	  SYM_OUT '=' NUM ';' {
		add_output(current_state, lex_sym_name, $3);
	}
	| IF '(' condition ')' NEXT SYM_STATE ';' {
		add_next_state(current_state, lex_sym_name, current_cond_list);
		current_cond_list = NULL;
	}
	| NEXT SYM_STATE ';' {
		add_next_state(current_state, lex_sym_name, NULL);
	}
	| NEXT SYM ';' {undef_state();}
	| SYM '=' NUM ';' {undef_output();}

condition:
	  cond
	| cond '&' '&' condition

cond:
	SYM_IN '=' '=' NUM {build_cond_list(lex_sym_name, $4);}

%%

void build_cond_list(const char *name, int value)
{
	struct cond_struct *new;

	if (value != 0 && value != 1) {
		fprintf(stderr,
			"Error, line %d, input condition for \"%s\" must be 0 or 1\n",
			lineno, name);
		err_count++;
		return;
	}
	new = (struct cond_struct *)malloc(sizeof(struct cond_struct));
	new->sym = lookup_symbol(name);
	if (new->sym == NULL) {
		fprintf(stderr, "Error, line %d, undefined input \"%s\"\n", 
			lineno, name);
		err_count++;
		free(new);
		return;
	}
	if (new->sym->type != TYPE_INPUT) {
		fprintf(stderr, "Error, line %d, \"%s\" is not an input\n", 
			lineno, name);
		err_count++;
		free(new);
		return;
	}
	new->value = value;
	new->next = current_cond_list;
	current_cond_list = new;
}



void add_symbol_group(int type, int value)
{
	int i;

	for (i=0; i<sym_count; i++) {
		add_symbol(sym_group[i], type, value);
	}
	sym_count = 0;
}


void undef_input(void)
{
	fprintf(stderr, "Error, line %d, Undefined input %s\n",
		lineno, lex_sym_name);
	err_count++;
}

void undef_output(void)
{
	fprintf(stderr, "Error, line %d, Undefined output %s\n",
		lineno, lex_sym_name);
	err_count++;
}

void undef_state(void)
{
	fprintf(stderr, "Error, line %d, Undefined state %s\n",
		lineno, lex_sym_name);
	err_count++;
}

void err_redefine(void)
{
	fprintf(stderr, "Error, line %d, Symbol %s redefined\n",
		lineno, lex_sym_name);
	err_count++;
}



int yyerror(char *s)
{
        if (yytext[0] >= 32) {
                fprintf(stderr, "%s near '%s', line %d\n",
                        s, yytext, lineno);
        } else {
                fprintf(stderr, "%s, line %d\n", s, lineno - 1);
        }
	err_count++;
        return 0;
}

int yywrap()
{
        return 1;
}

